import os

from batou.component import Component
from batou.lib.file import File


class PFADovecot(Component):

    local_conf = os.path.join(
        os.path.dirname(__file__), "dovecot", "local.conf"
    )
    database_conf = os.path.join(
        os.path.dirname(__file__), "dovecot", "database.conf"
    )

    def configure(self):
        self.db = self.require_one("pfa::database")
        self.keypair = self.require_one("keypair::mail")

        self += File("/etc/dovecot/local.conf", source=self.local_conf)
        self += File("database.conf", source=self.database_conf)



import os.path

import batou.component
import batou.lib.file

import batou_ext.nix
import batou_ext.ssl


@batou_ext.nix.rebuild
class Mirror(batou.component.Component):
    """Provide a package download mirror (server), and access to it (client).

    Usage: Server::

        [component:mirror]
        nginx_enable = True
        authstring = htpasswd-compatible file format
        nginx_config_path = /etc/local/nginx
        public_name = packages.example.com

    Usage: Client::

        [component:mirror]
        credentials = username:password
        public_name = packages.example.com


    Usage in components::

        mirror = self.require_one('mirror')
        self += Download(mirror.url('/path/to/package'))


    Server and client can be folded into one, when defining the options
    required for both in the same component.

    """

    public_name = None
    base = batou.component.Attribute(
        str, default="{protocol}://{credentials}@{public_name}"
    )
    protocol = batou.component.Attribute(str, default="https")

    # Whether Nginx-configuration should be provided
    nginx_enable = batou.component.Attribute("literal", default=False)
    nginx_config_path = None
    nginx_docroot = None

    # Define non-default reload command for Nginx used e.g. on renewal
    # of SSL-certificate if possible
    nginx_reload_command = batou.component.Attribute(
        str, default="sudo systemctl reload nginx"
    )

    provide_itself = batou.component.Attribute("literal", default=True)

    credentials = None
    authstring = None

    # Whether we shall run let's encrypt for Nginx configuration
    # If set to 'False', a self-signed certificate will be deployed instead
    use_letsencrypt = batou.component.Attribute("literal", default=True)

    def configure(self):

        if self.provide_itself:
            self.provide("mirror", self)

        if self.nginx_enable:
            self.nginx_enable_config()

    def url(self, path):
        return (
            self.base.format(
                protocol=self.protocol,
                credentials=self.credentials,
                public_name=self.public_name,
            )
            + "/"
            + path
        )

    def nginx_enable_config(self):
        self.address = batou.utils.Address(self.public_name, 80)
        self.ssl_address = batou.utils.Address(self.public_name, 443)

        if not self.nginx_docroot:
            self.nginx_docroot = self.map("htdocs")

        self += batou.lib.file.File(self.nginx_docroot, ensure="directory")

        self.cert = batou_ext.ssl.Certificate(
            self.public_name,
            docroot=self.nginx_docroot,
            extracommand=self.nginx_reload_command,
            use_letsencrypt=self.use_letsencrypt,
        )
        self += self.cert

        if self.authstring:
            self.htpasswdfile = batou.lib.file.File(
                "htpasswd", content=self.authstring
            )
            self += self.htpasswdfile

        assert self.nginx_config_path

        self += batou.lib.file.File(
            "{}/{}.conf".format(self.nginx_config_path, self.public_name),
            source=os.path.join(
                os.path.dirname(__file__), "resources/mirror.conf"
            ),
        )

        self += self.cert.activate_letsencrypt()



import batou
import pytest

from .. import oci


@pytest.fixture
def container(root):
    c = oci.Container(
        container_name="name",
        image="alpine",
    )

    c.prepare(root)
    return c


def patch_activate(mocker, activate):
    mocker.patch.object(activate, "_get_running_container_image_id")
    mocker.patch.object(activate, "_get_local_image_id")
    mocker.patch.object(activate, "_get_local_digest")
    mocker.patch.object(activate, "_docker_login")
    mocker.patch.object(activate, "_validate_remote_image")
    activate._validate_remote_image.return_value = False


@pytest.fixture
def activate(container, mocker, root):
    activate = container.activate()
    activate.prepare(root)
    patch_activate(mocker, activate)
    yield activate
    activate._remote_manifest_cache.clear()


def test_no_rebuild_flag_does_not_add_rebuild(root):
    c = oci.Container(
        container_name="name",
        image="alpine",
        rebuild=False,
    )
    c.prepare(root)
    with pytest.raises(AttributeError):
        c._activate_container

    # There are currently *no* component changes, as prepare only configures
    # the component, no verify has been called.
    assert not c.envfile.changed


def test_locally_updated_image_causes_restart(activate):
    activate._get_running_container_image_id.return_value = "v1"
    activate._get_local_image_id.return_value = "v2"

    with pytest.raises(batou.UpdateNeeded, match="Local version update."):
        activate.verify()
    assert activate._need_explicit_restart

    # There is no remote image validation in this case.
    activate._validate_remote_image.assert_not_called()


def test_remote_updated_image_causes_restart(activate):
    activate._get_running_container_image_id.return_value = "v1"
    activate._get_local_image_id.return_value = "v1"

    with pytest.raises(batou.UpdateNeeded, match="Remote version update."):
        activate.verify()
    assert activate._need_explicit_restart
    activate._validate_remote_image.assert_called()


def test_local_image_up_to_date_with_remote_causes_no_restart(activate):
    activate._get_running_container_image_id.return_value = "v1"
    activate._get_local_image_id.return_value = "v1"
    activate._validate_remote_image.return_value = True

    activate.verify()
    # Verify marks a restart *but* did not raise UpdatedNeede so update() is
    # never called!
    assert activate._need_explicit_restart

    # Remote image was validated:
    activate._validate_remote_image.assert_called()


def test_remote_image_validation_is_cached(root, activate, mocker):
    activate._get_running_container_image_id.return_value = "v1"
    activate._get_local_image_id.return_value = "v1"
    activate._validate_remote_image.return_value = False
    activate._get_local_digest.return_value = "local-digest"

    # Uncached call
    with pytest.raises(batou.UpdateNeeded, match="Remote version update."):
        activate.verify()

    # Validate is called with the local digest
    activate._validate_remote_image.assert_called_with(
        "alpine:latest@local-digest"
    )
    activate._validate_remote_image.reset_mock()

    # A *different* container will re-use the cache!
    c2 = oci.Container(container_name="second", image="alpine")
    c2.prepare(root)
    a2 = c2.activate()
    a2.prepare(root)

    patch_activate(mocker, a2)
    a2._get_running_container_image_id.return_value = "v1"
    a2._get_local_image_id.return_value = "v1"
    a2._validate_remote_image.return_value = False
    a2._get_local_digest.return_value = "local-digest"

    with pytest.raises(
        batou.UpdateNeeded, match="Cached remote version update."
    ):
        a2.verify()

    activate._validate_remote_image.assert_not_called()


def test_regsistry_address_triggers_login(container, activate):
    container.registry_address = "some-registry"
    activate._get_running_container_image_id.return_value = "v1"
    activate._get_local_image_id.return_value = "v1"
    with pytest.raises(batou.UpdateNeeded):
        activate.verify()
    activate._docker_login.assert_called_with()


def test_password_file_change_causes_restart(root):
    c = oci.Container(
        container_name="name",
        image="alpine",
        registry_address="mine",
        registry_user="ben",
        registry_password="utzer",
    )

    c.prepare(root)
    c.password_file.changed = True
    activate = c.activate()
    activate.prepare(root)

    with pytest.raises(batou.UpdateNeeded):
        activate.verify()

    assert activate._need_explicit_restart


def test_env_file_change_causes_restart(root):
    c = oci.Container(
        container_name="name",
        image="alpine",
    )

    c.prepare(root)
    c.envfile.changed = True
    activate = c.activate()
    activate.prepare(root)

    with pytest.raises(batou.UpdateNeeded):
        activate.verify()

    assert activate._need_explicit_restart



import os
import os.path
from glob import glob
from time import time

import batou.lib.archive
from batou.component import Attribute


class SingleUntar(batou.lib.archive.Untar):
    """Extract archive *once*

    This is useful when the target dir contains a version number or revision.
    In this case we do not expect that extracting the archive again could do
    any changes, so we skip the file-by-file validation which speeds things
    up considerably.

    """

    _required_params_ = {"target": "."}

    cleanup_markers_after_days = Attribute(int, default=None)

    def configure(self):
        super().configure()
        self._markers_to_unlink = []
        self._only_marker_cleanup = False
        self._finish_marker = os.path.join(
            os.path.dirname(self.target),
            f".{os.path.basename(self.target)}.finished-extract",
        )

    def verify(self):
        # If a deployment does this for a while, a lot of these markers accumulate.
        # If this attribute is set, cleanup all markers with an mtime older than
        # `cleanup_markers_after_days` days.
        if self.cleanup_markers_after_days:
            now = time()
            diff = now - self.cleanup_markers_after_days * 24 * 3600

            self._markers_to_unlink = [
                file
                for file in glob(self.map(".*.finished-extract"))
                if os.path.getmtime(file) < diff
            ]

        assert os.path.exists(self.archive)
        # Verify that the target dir is newer than the archive on disk.
        self.assert_file_is_current(self.target, [self.archive])
        # Verify that there is a finish marker and it is newer than the arvive
        self.assert_file_is_current(self._finish_marker, [self.archive])

        # If we got that far, we don't have to unpack again. But we should
        # clean up some old markers.
        self._only_marker_cleanup = True
        assert self._markers_to_unlink == []

    def update(self):
        if not self._only_marker_cleanup:
            super().update()
            self.touch(self._finish_marker)

        for old_marker in self._markers_to_unlink:
            os.remove(old_marker)



import batou.component


class NFS(batou.component.Component):
    """
    A component to help ensure you got access paths for NFS in sync
    for your deployment.

    Defaults are based on Flyingcircus' NixOS environment.
    """

    # Path where NFS on client is mounted on
    basepath = batou.component.Attribute(
        str, batou.component.ConfigString("/mnt/nfs/shared/")
    )

    # Path where NFS-share is located on the NFS server
    serverpath = batou.component.Attribute(
        str, batou.component.ConfigString("/srv/nfs/shared/")
    )

    def configure(self):
        self.provide("nfs", self)



import os
import shlex
from textwrap import dedent
from typing import Dict, Optional

import batou
from batou import UpdateNeeded
from batou.component import Attribute, Component
from batou.lib.file import File
from batou.utils import CmdExecutionError

import batou_ext.nix


class Container(Component):
    """A OCI Container component.

    With this component you can dynamically schedule docker containers to be
    run on the target host.

    Note: Docker image specifiers do not follow a properly resolvable pattern.
    Therefore, container registries have to be specified seperately if you need
    to log in before. If you do not provide a container registry, docker will
    use the default one for authentication. You can choose to also append the
    image attribute with the registry but this module will do so automatically.

    ```
    # the following two call are identical:
    self += batou_ext.oci.Container(
        image="foo",
        registry_address="test.registry",
        registry_user="foo",
        registry_password="bar")

    self += batou_ext.oci.Container(
        image="test.registry/foo",
        registry_address="test.registry",
        registry_user="foo",
        registry_password="bar")

    # However, this will fail since docker will try to log into the default container registry
    # with the provided credentials and then pull the image from the registry specified in the image string
    self += batou_ext.oci.Container(
        image="test.registry/foo",
        registry_user="foo",
        registry_password="bar")


    # if you don't need to authenticate, not explicitly specifying the registry is fine of course
    # and it will pull the image from the correct registry
    self += batou_ext.oci.Container(image="test.registry/foo")
    ```

    Example:
    ```
    self += batou_ext.oci.Container(image = "mysql", version = "8.0")
    ```

    If you have multiple containers on one machine you can consolidate the
    rebuild:

    ```
    self += batou_ext.oci.Container(
        image="mysql",
        version="8.0",
        rebuild=False,
    )
    container = self._
    # add more containers

    self += Rebuild()
    self += container.activate()




    """

    # general options
    image = Attribute(str)
    version: str = "latest"
    container_name = Attribute(str)

    # Set up monitoring
    monitor: bool = True

    # Automatically rebuild after this component
    rebuild: bool = True

    # specific options
    entrypoint: Optional[str] = None
    docker_cmd: Optional[str] = None
    envfile: Optional[File] = None
    mounts: dict = {}
    ports: dict = {}
    env: dict = {}
    depends_on: list = None
    extra_options: list = None

    # secrets
    registry_address = Attribute(Optional[str], None)
    registry_user = Attribute(Optional[str], None)
    registry_password = Attribute(Optional[str], None)

    _required_params_ = {
        "container_name": "mysql-container",
        "image": "mysql",
    }

    def configure(self):
        if (
            self.registry_user or self.registry_password
        ) and not self.registry_address:
            self.log(
                "WARN: you might want to specify the registry explicitly"
                " unless you really intend to log into the default"
                " docker registry"
            )

        if (
            self.registry_address
            # This is for the strange case of index.docker.io where you have
            # to set the registry_address to `https://index.docker.io/v1/`
            and not self.registry_address.startswith("https://")
            and not self.image.startswith(self.registry_address)
        ):
            self.image = f"{self.registry_address}{'/' if not self.registry_address.endswith('/') else ''}{self.image}"

        if self.registry_password:
            self += File(
                "registry-password", mode=0o600, content=self.registry_password
            )
            self.password_file = self._

        if self.envfile is None:
            self += File(
                f"{self.container_name}_env",
                sensitive_data=True,
                content="""{% for key, value in component.env.items() | sort -%}
{{key}}={{value}}
{% endfor %}""",
            )
            self.envfile = self._

        if self.docker_cmd:
            self._docker_cmd_list = shlex.split(self.docker_cmd)

        if not self.depends_on:
            self.depends_on = []

        self += File(
            f"/etc/local/nixos/docker_{self.container_name}.nix",
            sensitive_data=False,
            source=os.path.join(
                os.path.dirname(__file__), "resources/oci-template.nix"
            ),
        )

        if self.rebuild:
            self += batou_ext.nix.Rebuild()
            self += self.activate()
            self._activate_container = self._  # Test support

    def activate(self):
        """Return a component which actually activates the container."""
        return ContainerRestart(self)


class ContainerRestart(Component):
    """Helper component to restart container.

    This is separated from `Container` to be able to bundle the rebuild
    and subsequently restart the container.

    Note: a rebuild *must* happen after `Container` and before this component.

    """

    namevar = "container"

    # cache spanning multiple components deploying the same container
    # the values are bools indicating whether or not containers with
    # this specific digest are up to date
    _remote_manifest_cache: Dict[str, bool] = {}

    def verify(self):
        container = self.container

        # Only trigger restart if either file has been changed. A rebuild
        # will not restart the container.
        self._need_explicit_restart = container.envfile.changed or (
            container.registry_password and container.password_file.changed
        )
        container.assert_no_changes()
        container.envfile.assert_no_changes()

        # If we are past this point no sub components have been changed. If
        # an UpdateNeeded is raised we *do* need a restart
        self._need_explicit_restart = True

        # id of the running container
        container_image_id = self._get_running_container_image_id()

        # newest *locally* available id
        local_image_id = self._get_local_image_id()

        # If the container is not running the image we expect, we need to
        # restart it. This will also ensure the container is up-to-date locally
        # since the container's nix service is set to always pull.
        if local_image_id != container_image_id:
            self.log(
                "Container is running an older version than is locally "
                "available and needs to be restarted."
            )
            raise UpdateNeeded("Local version update.")

        # If the container is running the newest locally available image then
        # check if there is a newer version available remotely

        # query the local digest to compare against upstream
        local_digest = self._get_local_digest()

        image_ident = f"{container.image}:{container.version}@{local_digest}"

        # test whether the ident has been checked already
        if image_ident in self._remote_manifest_cache:
            if self._remote_manifest_cache[image_ident]:
                # the image is up to date -> does not need an update
                return
            raise UpdateNeeded("Cached remote version update.")

        if container.registry_address:
            self._docker_login()

        valid = self._validate_remote_image(image_ident)

        # add the validated ident to the cache so that components using the
        # same container # and same versions don't have to query the remote
        self._remote_manifest_cache[image_ident] = valid
        if not valid:
            raise UpdateNeeded("Remote version update.")

    def update(self):
        if self._need_explicit_restart:
            self.cmd(
                f"sudo systemctl restart docker-{self.container.container_name}"
            )

    def _get_running_container_image_id(self):
        """Get the image Id the container is currently running.

        If there is no container, return "null".
        """
        image_id, stderr = self.cmd(
            dedent(
                """\
            docker container inspect {{component.container.container_name}} \
                | jq -r '.[0].Image'
                """
            )
        )
        return image_id

    def _get_local_image_id(self):
        """Return the id of the image we have downloaded.

        If there is no image, return "null" .
        """
        local_image_id, stderr = self.cmd(
            dedent(
                """\
            docker image inspect {{component.container.image}}:{{component.container.version}} \
                | jq -r '.[0].Id' \
                || echo image not available locally
                """
            )
        )
        return local_image_id

    def _get_local_digest(self):
        local_digest, stderr = self.cmd(
            dedent(
                """\
            docker image inspect {{component.container.image}}:{{component.container.version}} \
                | jq -r 'first | .RepoDigests | first | split("@") | last' \
                || echo image not available locally
                """
            )
        )
        return local_digest

    def _docker_login(self):
        self.cmd(
            self.expand(
                dedent(
                    """\
        docker login \\
            {%- if component.container.registry_user and component.container.registry_password %}
            -u {{component.container.registry_user}} \\
            -p {{component.container.registry_password}} \\
            {%- endif %}
            {{component.container.registry_address}}
        """
                )
            )
        )

    def _validate_remote_image(self, image_ident):
        try:
            # check if the digest aligns with the remote image
            # if it does not, this command will throw an error
            stdout, stderr = self.cmd(f"docker manifest inspect {image_ident}")
        except CmdExecutionError as e:
            error = e.stderr
            if error.startswith("unsupported manifest format"):  # gitlab
                error = (
                    "Local and remote digest are out of sync! "
                    "The container needs to be restarted."
                )
            batou.output.annotate(error, debug=True)
            valid = False
        else:
            # `docker manifest inspect` silently raises an error when unauthorized,
            # returns exit code 0
            if stderr == "unauthorized":
                raise RuntimeError(
                    "Wrong credentials for remote container registry"
                )
            valid = True
        return valid



from .mail import Mailhog


class Mailhog(Mailhog):
    """(deprecated) BW compatibility for .mail.Mailhog."""

    def configure(self):
        super().configure()
        self.address_mail = self.address
        self.address = self.address_ui



from textwrap import dedent

import batou.component
import batou.lib.nagios
import pkg_resources


class CronJob(batou.component.Component):
    """Improved cronjob with locking, timeout, logging, and monitoring

    See https://blog.flyingcircus.io/2015/04/16/improving-periodic-data-import-jobs-in-3-steps/
    for motivation etc.

    Usage::

        self += batou_ext.cron.CronJob(
            "cleanup",
            command="bin/cleanup some stuff",
            timing="*/4 * * * * ",
            timeout="3m",
            checkWarning=6,
            checkCritical=10)

    Where `tag` is an identifier from which is derived: script name,
    lock file, log file, stamp file.

    `checkWarning` and `checkCritical` should be given in minutes since
    last update of the stamp-file (=last successfully run of cronJob)
    and will turn the service check into warning or critical state after
    these times were exceeded.
    Ensure this values are fitting the settings done within `timing`.

    Please note, that this component requires you to import the
    `Logrotate` component (`from batou.lib.logrotate import Logrotate`)
    and add it to the components list of the host on which you want to
    deploy the cronjob.

    e.g.

            [host:yourhost]
                components =
                    ...
                    logrotate
    """  # noqa: E501 line too long

    _required_params_ = {"timing": "*/4 * * * * "}
    namevar = "tag"

    command = None
    timing = None
    log_file = None
    lock_file = None
    stamp_file = None
    timeout = "1h"
    checkWarning = None  # minutes
    checkCritical = None  # minutes

    args = ""  # to satisfy sorting

    def format(self):
        return self.expand(
            """\
{{component.timing}} \
 timeout {{component.timeout}} \
 {{component.wrapped_command}} \
 >> {{component.log_file}} 2>&1"""
        )

    def configure(self):
        self.provide(batou.lib.cron.CronJob.key, self)

        if self.timing is None:
            raise ValueError(
                "Required timing value missing from cron job %r." % self.command
            )

        self += batou.lib.file.File("logs", ensure="directory")
        self.log_file = self.map(self.expand("logs/{{component.tag}}.log"))
        self += batou.lib.logrotate.RotatedLogfile(self.log_file)

        self.lock_file = self.map(self.expand(".{{component.tag}}.lock"))
        self.stamp_file = self.map(self.expand(".{{component.tag}}.stamp"))

        # Ensure, we do have the stamp file there so the sensu-check is
        # getting red if next runs are not successful
        self += batou.lib.file.File(self.stamp_file, content="")

        self += batou.lib.file.File(
            self.expand("{{component.tag}}.sh"),
            content=pkg_resources.resource_string(
                __name__, "resources/cron-wrapper.sh"
            ),
            mode=0o755,
        )
        self.wrapped_command = self._.path

        if self.checkWarning or self.checkCritical:
            cmd = ["-f", self.stamp_file]
            if self.checkWarning:
                cmd.append("-w")
                cmd.append(str(int(self.checkWarning) * 60))
            if self.checkCritical:
                cmd.append("-c")
                cmd.append(str(int(self.checkCritical) * 60))

            self += batou.lib.nagios.Service(
                self.expand("Cronjob {{component.tag}} finished?"),
                command="check_file_age",
                args=" ".join(cmd),
                name=self.expand("cronjob_{{component.tag}}"),
            )


class SystemdTimer(batou.component.Component):
    """Integration for systemd timers with *FC nixos*.


    Usage::

        self += batou_ext.cron.SystemdTimer(
            "cleanup",
            command="bin/cleanup some stuff",
            onCalendar="*:0/5",
            timeout="3m")

    See https://www.freedesktop.org/software/systemd/man/systemd.time.html
    for details about the `onCalendar` timer settings.

    If `persistent` is set to True, the time when the service unit was last
    triggered is stored on disk.
    When the timer is activated, the service unit is triggered immediately
    if it would have been triggered at least once during the time when
    the timer was inactive.

    For deployments, this usually means that the unit is started immediately
    after, without waiting for the next scheduled activation.
    """

    _required_params_ = {"command": "/bin/true", "onCalendar": "02:00:00"}

    namevar = "tag"
    command = batou.component.Attribute(str)
    onCalendar = batou.component.Attribute(str)
    persistent = batou.component.Attribute("literal", False)
    timeout = "1h"
    description = None
    additional_service_config = None
    run_as = batou.component.Attribute(
        str,
        default=batou.component.ConfigString(
            "{{component.environment.service_user}}"
        ),
    )

    def configure(self):
        if self.description is None:
            self.description = f"Batou generated unit for {self.tag}"

        self += batou.lib.file.File(
            self.expand("{{component.tag}}.sh"),
            mode=0o755,
            content=dedent(
                f"""\
                #!/bin/sh

                # This file is generated by batou and run by systemd
                # via `{self.tag}.timer` and `{self.tag}.service`.

                source /etc/profile
                set -exo pipefail

            """
            )
            + self.command,
        )
        self.wrapped_command = self._.path

        if isinstance(self.persistent, bool) and self.persistent:
            self.persistent_timer_config = "true"
        else:
            self.persistent_timer_config = "false"

        self += batou.lib.file.File(
            f"/etc/local/nixos/timer-{self.tag}.nix",
            content=dedent(
                """\
              { ... }:
              {
                systemd.timers."{{component.tag}}" = {
                  wantedBy = [ "timers.target" ];
                  timerConfig = {
                    OnCalendar = "{{component.onCalendar}}";
                    Persistent = {{component.persistent_timer_config}};
                  };
                };

                systemd.services."{{component.tag}}" = {
                  description = "{{component.description}}";
                  serviceConfig = {
                    Type = "oneshot";
                    User = "{{component.run_as}}";
                    ExecStart = "{{component.wrapped_command}}";
                    TimeoutStartSec = "{{component.timeout}}";
                    {%- if component.additional_service_config is not none %}
                    {%- for setting, value in component.additional_service_config.items() %}
                    {{ setting }} = {{ value }};
                    {%- endfor %}
                    {%- endif %}
                  };
                };
              }
        """
            ),
        )

        self += batou.lib.file.File(
            self.expand("check_systemd_unit.py"),
            content=pkg_resources.resource_string(
                __name__, "resources/check_systemd_unit.py"
            ),
            mode=0o755,
        )
        self.check_command = self._.path
        self += batou.lib.nagios.Service(
            f"Timer/Unit {self.tag} ok?",
            command=self.check_command,
            args=self.tag,
            name=f"timer_{self.tag}",
        )
